import os
from os.path import join
from typing import Optional, Tuple

import numpy as np
import pandas as pd
from botocore.model import InvalidShapeError
from numpy._typing import ArrayLike

from psf_analysis_CFIM.debug.debug import report_error_debug
from psf_analysis_CFIM.error_widget.error_display_widget import report_error
from psf_analysis_CFIM.psf_analysis.extract.BeadExtractor import BeadExtractor
from psf_analysis_CFIM.psf_analysis.image import Calibrated3DImage
from psf_analysis_CFIM.psf_analysis.parameters import PSFAnalysisInputs
from psf_analysis_CFIM.psf_analysis.psf import PSF


class Analyzer:
    def __init__(self, parameters: PSFAnalysisInputs, analyzer_settings: dict, info_dict=None):
        if info_dict is None:
            info_dict = {"wavelength_color": "black",
                        "wavelength": "",
                        "excitation": "",
                        "airy_unit": "", }
        self._parameters = parameters
        self._settings = analyzer_settings
        bead_extractor = BeadExtractor(
            image=Calibrated3DImage(
                data=parameters.img_data, spacing=parameters.spacing
            ),
            patch_size=parameters.patch_size,
        )
        self._bead_margins = bead_extractor.get_margins()
        self._airy_unit = info_dict["airy_unit"]
        self._excitation = info_dict["excitation"]
        self._wavelength = info_dict["wavelength"]
        self._wavelength_color = info_dict["wavelength_color"]
        self._invalid_beads_index = []
        self._beads = bead_extractor.extract_beads(points=self._parameters.point_data)


        self._extractor_points_diff = len(self._parameters.point_data) - len(self._beads)
        self._results = None
        self._result_figures = {}
        self._index = 0

        self._debug = self._settings.get("debug")

        if self._debug:
            print(f"Debug | Analyzer for {self._wavelength} with {len(self._beads)} beads")

    def __iter__(self):
        return self

    def __len__(self):
        return len(self._beads)

    def __next__(self):
        if self._index < len(self._beads):
            bead = self._beads[self._index]
            try:
                if bead.data.shape != (int(self._bead_margins[0]), int(self._bead_margins[1]), int(self._bead_margins[2])):
                    raise InvalidShapeError(f"Discarding bead with invalid shape: {bead.data.shape}")
                if 0 in bead.data.shape:
                    raise InvalidShapeError(f"Discarding bead with invalid shape: {bead.data.shape}")
                psf = PSF(image=bead, psf_settings=self._settings["psf_settings"])
                psf.analyze()
                if psf.error:
                    raise InvalidShapeError(f"Discarding bead due to analyze error: {self._index}")
                results = psf.get_summary_dict()
                extended_results = self._extend_result_table(bead, results)
                self._add(
                    extended_results,
                    psf.get_summary_image(
                        date=self._parameters.date,
                        version=self._parameters.version,
                        dpi=self._parameters.dpi,
                        ellipsoid_color=self._wavelength_color,
                        centroid= (extended_results["z_mu"], extended_results["y_mu"], extended_results["x_mu"]),
                    ),
                )
            except InvalidShapeError as e:
                self._invalid_beads_index.append(self._index)
                # min_cord, max_cord = bead.get_box()
                report_error("", bead.get_middle_coordinates())

            self._index += 1
            return self._index + self._extractor_points_diff, self._wavelength_color
        else:
            if self._debug:
                print(f"Analyzer {self._wavelength}| Finished analyzing {len(self._beads)} beads")
            raise StopIteration()

    def get_date(self):
        """Return the date of the image."""
        return self._parameters.date

    def get_version(self):
        """Return the version of the analysis."""
        return self._parameters.version

    def get_dpi(self):
        """Return the DPI of the analysis."""
        return self._parameters.dpi

    def get_wavelength(self):
        """Return the wavelength of the image."""
        return self._wavelength

    def get_wavelength_color(self):
        """Return the color of the wavelength."""
        return self._wavelength_color

    def get_centroids(self):
        """Return the centroids of the beads."""
        centroids = []
        for i in range(len(self._results["x_mu"])):
            centroids.append((int(self._results["z_mu"][i]), int(self._results["y_mu"][i]), int(self._results["x_mu"][i])))
        return centroids


    def get_raw_beads(self):
        """Return the raw data of the beads before analysis."""
        return self._beads

    def get_raw_beads_filtered(self):
        """ Return the raw data of the beads before analysis, but filtered for found invalid shapes."""
        beads = [bead for i, bead in enumerate(self._beads) if i not in self._invalid_beads_index]
        return beads

    def get_averaged_bead(self):
        """Average the raw bead data before analysis."""
        filtered_beads = [bead.data for bead in self.get_raw_beads_filtered()]
        try:
            averaged_bead_data = np.mean(filtered_beads, axis=0).astype(np.uint16)
            averaged_bead = Calibrated3DImage(data=averaged_bead_data, spacing=self._parameters.spacing)
            return averaged_bead
        except (ValueError, AttributeError) as e:
            print(f"Error getting average bead of {len(filtered_beads)} beads: {e}")
            if self._debug:
                report_error_debug(filtered_beads, "3d_array")



    def _extend_result_table(self, bead, results):
        extended_results = results.copy()
        extended_results["z_mu"] += bead.offset[0] * self._parameters.spacing[0]
        extended_results["y_mu"] += bead.offset[1] * self._parameters.spacing[1]
        extended_results["x_mu"] += bead.offset[2] * self._parameters.spacing[2]
        extended_results["image_name"] = self._parameters.name
        extended_results["date"] = self._parameters.date
        extended_results["microscope"] = self._parameters.microscope
        extended_results["mag"] = self._parameters.magnification
        extended_results["NA"] = self._parameters.na
        extended_results["yx_spacing"] = self._parameters.spacing[1]
        extended_results["z_spacing"] = self._parameters.spacing[0]
        extended_results["version"] = self._parameters.version
        return extended_results

    def _add(self, result: dict, summary_fig: ArrayLike):
        if self._results is None:
            self._results = {}
            for key in result.keys():
                self._results[key] = [result[key]]
            self._results["PSF_path"] = []
        else:
            for key in result.keys():
                self._results[key].append(result[key])

        centroid = np.round([result["x_mu"], result["y_mu"], result["z_mu"]], 1)
        bead_name = "{}_Bead_X{}_Y{}_Z{}".format(result["image_name"], *centroid)
        unique_bead_name = self._make_unique(bead_name)
        self._result_figures[unique_bead_name] = summary_fig
        self._results["PSF_path"].append(join(unique_bead_name + ".png"))

    def _make_unique(self, name: str):
        count = 1
        unique_name = name
        while unique_name in self._result_figures.keys():
            unique_name = f"{name}-({count})"
            count += 1
        return unique_name

    def get_results(self) -> Optional[pd.DataFrame]:
        """Create result table from dict.

        Parameters
        ----------
        results :
            Result dict obtained from `analyze_bead`

        Returns
        -------
        result_table
            Result table with "nice" column names
        """
        if self._results is not None:
            return self._build_dataframe()
        else:
            return None

    def get_summary_figure_stack(
        self,
        bead_img_scale: Tuple[float, float, float],
        bead_img_shape: Tuple[int, int, int],
    ) -> Optional[Tuple[ArrayLike, ArrayLike]]:
        """
        Create a (N, Y, X, 3) stack of all summary figures.

        Parameters
        ----------
        bead_img_scale : scaling of the whole bead image in (Z, Y, X) nm/px
        bead_img_shape : shape of the whole bead image

        Returns
        -------
        stack of all summary figures
        scaling to display them with napari
        """
        if len(self._result_figures) > 0:
            measurement_stack = self._build_figure_stack()
            measurement_scale = self._compute_figure_scaling(
                bead_img_scale, bead_img_shape, measurement_stack
            )

            return measurement_stack, measurement_scale
        else:
            return None, None

    def _compute_figure_scaling(self, bead_scale, bead_shape, measurement_stack):
        measurement_scale = np.array(
            [
                bead_scale[0],
                bead_scale[1] / measurement_stack.shape[1] * bead_shape[1],
                bead_scale[1] / measurement_stack.shape[1] * bead_shape[1],
            ]
        )
        return measurement_scale

    def _build_figure_stack(self):
        measurement_stack = np.stack(
            [self._result_figures[k] for k in self._result_figures.keys()]
        )
        return measurement_stack

    def  _build_dataframe(self):
        return pd.DataFrame(
            {
                "ImageName": self._results["image_name"],
                "Date": self._results["date"],
                "Microscope": self._results["microscope"],
                "Magnification": self._results["mag"],
                "NA": self._results["NA"],
                "Emission": self._wavelength,
                "Excitation": self._excitation,
                "AiryUnit": self._airy_unit,
                "Amplitude_1D_Z": self._results["z_amp"],
                "Amplitude_2D_XY": self._results["yx_amp"],
                "Amplitude_3D_XYZ": self._results["zyx_amp"],
                "Background_1D_Z": self._results["z_bg"],
                "Background_2D_XY": self._results["yx_bg"],
                "Background_3D_XYZ": self._results["zyx_bg"],
                "Z_1D": self._results["z_mu"],
                "X_2D": self._results["x_mu"],
                "Y_2D": self._results["y_mu"],
                "X_3D": self._results["zyx_x_mu"],
                "Y_3D": self._results["zyx_y_mu"],
                "Z_3D": self._results["zyx_z_mu"],
                "FWHM_1D_Z": self._results["z_fwhm"],
                "FWHM_2D_X": self._results["x_fwhm"],
                "FWHM_2D_Y": self._results["y_fwhm"],
                "FWHM_3D_Z": self._results["zyx_z_fwhm"],
                "FWHM_3D_Y": self._results["zyx_y_fwhm"],
                "FWHM_3D_X": self._results["zyx_x_fwhm"],
                "FWHM_PA1_2D": self._results["yx_pc1_fwhm"],
                "FWHM_PA2_2D": self._results["yx_pc2_fwhm"],
                "FWHM_PA1_3D": self._results["zyx_pc1_fwhm"],
                "FWHM_PA2_3D": self._results["zyx_pc2_fwhm"],
                "FWHM_PA3_3D": self._results["zyx_pc3_fwhm"],
                "SignalToBG_1D_Z": (
                    np.array(self._results["z_amp"]) / np.array(self._results["z_bg"])
                ).tolist(),
                "SignalToBG_2D_XY": (
                    np.array(self._results["yx_amp"]) / np.array(self._results["yx_bg"])
                ).tolist(),
                "SignalToBG_3D_XYZ": (
                    np.array(self._results["zyx_amp"])
                    / np.array(self._results["zyx_bg"])
                ).tolist(),
                "xy_pixelsize": self._results["yx_spacing"],# Changed these two from XYspacing and Zspacing
                "z_spacing": self._results["z_spacing"],
                "cov_xx_3D": self._results["zyx_cxx"],
                "cov_xy_3D": self._results["zyx_cyx"],
                "cov_xz_3D": self._results["zyx_czx"],
                "cov_yy_3D": self._results["zyx_cyy"],
                "cov_yz_3D": self._results["zyx_czy"],
                "cov_zz_3D": self._results["zyx_czz"],
                "cov_xx_2D": self._results["yx_cxx"],
                "cov_xy_2D": self._results["yx_cyx"],
                "cov_yy_2D": self._results["yx_cyy"],
                "sde_amp_1D_Z": self._results["z_amp_sde"],
                "sde_amp_2D_XY": self._results["yx_amp_sde"],
                "sde_amp_3D_XYZ": self._results["zyx_amp_sde"],
                "sde_background_1D_Z": self._results["z_bg_sde"],
                "sde_background_2D_XY": self._results["yx_bg_sde"],
                "sde_background_3D_XYZ": self._results["zyx_bg_sde"],
                "sde_Z_1D": self._results["z_mu_sde"],
                "sde_X_2D": self._results["x_mu_sde"],
                "sde_Y_2D": self._results["y_mu_sde"],
                "sde_X_3D": self._results["zyx_x_mu_sde"],
                "sde_Y_3D": self._results["zyx_y_mu_sde"],
                "sde_Z_3D": self._results["zyx_z_mu_sde"],
                "sde_cov_xx_3D": self._results["zyx_cxx_sde"],
                "sde_cov_xy_3D": self._results["zyx_cyx_sde"],
                "sde_cov_xz_3D": self._results["zyx_czx_sde"],
                "sde_cov_yy_3D": self._results["zyx_cyy_sde"],
                "sde_cov_yz_3D": self._results["zyx_czx_sde"],
                "sde_cov_zz_3D": self._results["zyx_czz_sde"],
                "sde_cov_xx_2D": self._results["yx_cxx_sde"],
                "sde_cov_xy_2D": self._results["yx_cyx_sde"],
                "sde_cov_yy_2D": self._results["yx_cyy_sde"],
                "z_pos": self._results["z_mu"],
                "y_pos": self._results["y_mu"],
                "x_pos": self._results["x_mu"],
                "version": self._results["version"],
                "PSF_path": self._results["PSF_path"],
            }
        )



# File: tests/fake_viewer.py
from psf_analysis_CFIM.tests.fake_event import FakeEvents
from psf_analysis_CFIM.tests.fake_layer_list import FakeLayerList


# A fake viewer with a layers attribute and events.
class FakeViewer:
    def __init__(self, layers):
        self.layers = FakeLayerList(layers)
        self.events = FakeEvents()

    def add_layer(self, layer):
        self.layers._layers.append(layer)
        self.events.layers_change.emit()


import os
from typing import overload

import numpy as np
from napari.utils.theme import opacity
from qtpy.QtCore import QObject, Signal
from scipy import spatial
from vispy.visuals import LineVisual

from psf_analysis_CFIM.debug import global_vars


class DebugEmitter(QObject):
    save_debug = Signal(object, str)

def report_error_debug(message, msg_type="message"):
    debug_emitter.save_debug.emit(message, msg_type)

debug_emitter = DebugEmitter()

# TODO: Flush out the potential features for dedicated debug class
class DebugClass:
    """
        Holds methods for debugging and visualizing data and processes.

        Is instantiated in napari console when PSF_ANALYSIS_CFIM_DEBUG == "1"
        Because of this, all methods should be callable from the console.

    """
    # region Initialization
    def __init__(self, viewer):
        self.errors: list[tuple[any, str]] = []
        self.error_amount = 0
        self._viewer = viewer
        self._widget = None
        self._bead_layer = None
        self._shape_layer = None
        self._surface_layer = None

        global_vars.debug_instance = self

        debug_emitter.save_debug.connect(self.save_debug)

    def set_PSFAnalysis_instance(self, widget):
        self._widget = widget

    def save_debug(self, message, msg_type):
        self.error_amount += 1
        self.errors.append((message, msg_type))

    # endregion
    # region Visualization | 3D

    def show_3d_array(self, index):
        self._viewer.add_image(self.get(index))

    @overload
    def add_nm_point(self, point: tuple[float, float, float]):
        pass

    @overload
    def add_nm_point(self, z: float, y: float, x: float):
        pass


    def add_nm_point(self, *args):
        if len(args) == 1:
            point = args[0]
        else:
            point = (args[0], args[1], args[2])
        scaled_point = point / self._widget.get_scale()
        self._viewer.add_points(scaled_point, size=10, face_color='violet', name="nm_point", scale=self._widget.get_scale())

    def show_psf_box(self, index):
        if self._set_bead_point_layer():
            point = self._bead_layer.data[index]
            self._add_bounding_box(point)

    def _add_bounding_box(self, point: tuple[float, float, float], *, bounding_box_dims=None):
        if bounding_box_dims is None:
            bounding_box_dims = self._widget.get_bounding_box_px()

        # surface_layer = self._get_surface_layer()

        half_bbox = (bounding_box_dims[0] / 2, bounding_box_dims[1] / 2, bounding_box_dims[2] / 2)

        # Adjust vertices to be centered around the given point and scale them
        vertices = np.array([
            [point[0] - half_bbox[0], point[1] - half_bbox[1], point[2] - half_bbox[2]],
            [point[0] - half_bbox[0], point[1] - half_bbox[1], point[2] + half_bbox[2]],
            [point[0] - half_bbox[0], point[1] + half_bbox[1], point[2] - half_bbox[2]],
            [point[0] - half_bbox[0], point[1] + half_bbox[1], point[2] + half_bbox[2]],
            [point[0] + half_bbox[0], point[1] - half_bbox[1], point[2] - half_bbox[2]],
            [point[0] + half_bbox[0], point[1] - half_bbox[1], point[2] + half_bbox[2]],
            [point[0] + half_bbox[0], point[1] + half_bbox[1], point[2] - half_bbox[2]],
            [point[0] + half_bbox[0], point[1] + half_bbox[1], point[2] + half_bbox[2]]
        ]) * self._widget.get_scale()

        hull = spatial.ConvexHull(vertices)
        surface = (hull.points, hull.simplices)

        if self._surface_layer is None:
            self._surface_layer = self._viewer.add_surface(surface, opacity=0, name="bbox_surface", wireframe={'visible': True, 'color': 'red'})
        else:
            self._surface_layer.data = surface
        # else: # TODO: Figure out how to add multiple surfaces to the same layer
        #     data = self._surface_layer.data
        #     data = np.array(data[0])
        #     data = np.concatenate((data, surface), axis=0)
        #     self._surface_layer.data = data


    def _draw_bounding_box(self, min_coord, max_coord):
        """
            Returns a list of line visuals representing the edges of a bounding box.
        """
        box = [
            [min_coord, [min_coord[0], min_coord[1], max_coord[2]]],
            [min_coord, [min_coord[0], max_coord[1], min_coord[2]]],
            [min_coord, [max_coord[0], min_coord[1], min_coord[2]]],
            [max_coord, [max_coord[0], max_coord[1], min_coord[2]]],
            [max_coord, [max_coord[0], min_coord[1], max_coord[2]]],
            [max_coord, [min_coord[0], max_coord[1], max_coord[2]]],
            [[max_coord[0], min_coord[1], min_coord[2]], [max_coord[0], max_coord[1], min_coord[2]]],
            [[max_coord[0], min_coord[1], min_coord[2]], [max_coord[0], min_coord[1], max_coord[2]]],
            [[min_coord[0], max_coord[1], min_coord[2]], [min_coord[0], max_coord[1], max_coord[2]]],
            [[min_coord[0], min_coord[1], max_coord[2]], [min_coord[0], max_coord[1], max_coord[2]]],
            [[min_coord[0], min_coord[1], max_coord[2]], [max_coord[0], min_coord[1], max_coord[2]]],
            [[min_coord[0], max_coord[1], min_coord[2]], [max_coord[0], max_coord[1], min_coord[2]]]
        ]
        return box

    def _draw_vertices_box(self, min_coord, max_coord):
        # vertices are generated according to the following scheme:
        #    5-------7
        #   /|      /|
        #  1-------3 |
        #  | |     | |
        #  | 4-----|-6
        #  |/      |/
        #  0-------2
        vertices = [
            min_coord,
            [max_coord[0], min_coord[1], min_coord[2]],
            [min_coord[0], min_coord[1], max_coord[2]],
            [max_coord[0], min_coord[1], max_coord[2]],
            [min_coord[0], max_coord[1], min_coord[2]],
            [max_coord[0], max_coord[1], min_coord[2]],
            [min_coord[0], max_coord[1], max_coord[2]],
            max_coord
        ]
        return vertices


    def _add_wireframe(self, min_coord, max_coord):
        """
            Add a wireframe box to the viewer.

            Parameters:
                min_coord (tuple): The minimum coordinates of the box (x, y, z)
                max_coord (tuple): The maximum coordinates of the box (x, y, z)
        """

        # Define the 8 vertices of the box using the min and max coordinates
        vertices = np.array([
            [min_coord[0], min_coord[1], min_coord[2]],
            [max_coord[0], min_coord[1], min_coord[2]],
            [max_coord[0], max_coord[1], min_coord[2]],
            [min_coord[0], max_coord[1], min_coord[2]],
            [min_coord[0], min_coord[1], max_coord[2]],
            [max_coord[0], min_coord[1], max_coord[2]],
            [max_coord[0], max_coord[1], max_coord[2]],
            [min_coord[0], max_coord[1], max_coord[2]],
        ])

        # Define the edges of the box as pairs of vertex indices
        edges = [
            [0, 1], [1, 2], [2, 3], [3, 0],  # bottom face
            [4, 5], [5, 6], [6, 7], [7, 4],  # top face
            [0, 4], [1, 5], [2, 6], [3, 7]  # vertical edges
        ]

        # Create segments for each edge (each segment is a pair of coordinates)
        segments = np.array([[vertices[i], vertices[j]] for i, j in edges])


        # Add the segments as a shapes layer (using 'line' type for wireframe)
        self._viewer.add_shapes(
            segments,
            shape_type='line',
            edge_color='red',
            face_color='transparent',
            scale=self._scale
        )
    # endregion
    # region Helper methods
    def get(self, index):
        return self.errors[index]

    def _get_shape_layer(self):
        layer_name = "Shapes"
        scale = self._widget.get_scale()

        if self._shape_layer is None:
            self._shape_layer = self._viewer.add_shapes(np.empty((0, 3)), ndim=3, name=layer_name, scale=scale)

        return self._shape_layer

    def _get_shape_points_layer(self):
        layer_name = "Shape Points"
        scale = self._widget.get_scale()

        if self._shape_points_layer is None:
            self._shape_points_layer = self._viewer.add_points(np.empty((0, 3)), ndim=3, name=layer_name, scale=scale
                                                               ,face_color='blue', border_color = "blue" ,size=2)

        return self._shape_points_layer

    def _get_surface_layer(self):
        layer_name = "bbox_surface"

        if self._shape_layer is None:
            self._shape_layer = self._viewer.add_surface(np.empty((0, 3)), name=layer_name)

        return self._shape_layer

    def _set_bead_point_layer(self):
        """
            Gets the selected bead layer from PSFAnalysis widget and sets it to self._bead_layer
        """
        layer = self._widget.get_current_points_layer()
        if layer is None:
            print("No bead layer selected.")
            return None
        if layer == self._bead_layer:
            return True
        self._bead_layer = layer
        return True

    def say_hello(self):
        return "Hello from Debug!"
    # endregion

    # region tests
    def test_type(self, index):
        # Language: python
        error = self.get(0)  # Get the error record
        data = error[0]       # Data to be saved

        print("Type of data:", type(data))
        if hasattr(data, "shape"):
            print("Shape of data:", data.shape)
        else:
            print("Data does not have a shape attribute")

        # Optionally, print first few elements if it is a sequence
        if hasattr(data, "__iter__"):
            print("Data sample:", list(data)[:5])
        else:
            print("Data is not iterable.")

    # endregion
    def save(self, index, *, overwrite_type=None):
        output_folder = "saved_debug"
        error = self.get(index)
        data =  error[0]
        msg_type = overwrite_type if overwrite_type else error[1] # One-liner for overwrite

        # Make directory if needed
        os.makedirs(output_folder, exist_ok=True)

        if msg_type == "3d_array":
            np.save(f"{output_folder}/debug_{index}.npy", error[0])
        elif msg_type == "message":
            with open(f"{output_folder}/debug_{index}.txt", "w") as f:
                f.write(str(data))
        else:
            print(f"Unknown message type: {msg_type}")





from typing import Tuple

from pydantic import BaseModel, NonNegativeFloat, PositiveFloat, PositiveInt


class ZFitRecord(BaseModel):
    z_bg: float
    z_amp: float
    z_mu: NonNegativeFloat
    z_sigma: NonNegativeFloat
    z_fwhm: NonNegativeFloat
    z_bg_sde: NonNegativeFloat
    z_amp_sde: NonNegativeFloat
    z_mu_sde: NonNegativeFloat
    z_sigma_sde: NonNegativeFloat


class YXFitRecord(BaseModel):
    yx_bg: NonNegativeFloat
    yx_amp: NonNegativeFloat
    y_mu: NonNegativeFloat
    x_mu: NonNegativeFloat
    yx_cyy: float
    yx_cyx: float
    yx_cxx: float
    y_fwhm: NonNegativeFloat
    x_fwhm: NonNegativeFloat
    yx_pc1_fwhm: NonNegativeFloat
    yx_pc2_fwhm: NonNegativeFloat
    yx_bg_sde: NonNegativeFloat
    yx_amp_sde: NonNegativeFloat
    y_mu_sde: NonNegativeFloat
    x_mu_sde: NonNegativeFloat
    yx_cyy_sde: NonNegativeFloat
    yx_cyx_sde: NonNegativeFloat
    yx_cxx_sde: NonNegativeFloat


class ZYXFitRecord(BaseModel):
    zyx_bg: NonNegativeFloat
    zyx_amp: NonNegativeFloat
    zyx_z_mu: float
    zyx_y_mu: float
    zyx_x_mu: float
    zyx_czz: float
    zyx_czy: float
    zyx_czx: float
    zyx_cyy: float
    zyx_cyx: float
    zyx_cxx: float
    zyx_z_fwhm: NonNegativeFloat
    zyx_y_fwhm: NonNegativeFloat
    zyx_x_fwhm: NonNegativeFloat
    zyx_pc1_fwhm: NonNegativeFloat
    zyx_pc2_fwhm: NonNegativeFloat
    zyx_pc3_fwhm: NonNegativeFloat
    zyx_bg_sde: NonNegativeFloat
    zyx_amp_sde: NonNegativeFloat
    zyx_z_mu_sde: NonNegativeFloat
    zyx_y_mu_sde: NonNegativeFloat
    zyx_x_mu_sde: NonNegativeFloat
    zyx_czz_sde: NonNegativeFloat
    zyx_czy_sde: NonNegativeFloat
    zyx_czx_sde: NonNegativeFloat
    zyx_cyy_sde: NonNegativeFloat
    zyx_cyx_sde: NonNegativeFloat
    zyx_cxx_sde: NonNegativeFloat


class PSFRecord(BaseModel):
    z_fit: ZFitRecord
    yx_fit: YXFitRecord
    zyx_fit: ZYXFitRecord


class PSFAnalysisInputs(BaseModel):
    date: Tuple[int, int, int]
    microscope: str
    magnification: PositiveInt
    objective_id: str
    na: PositiveFloat
    spacing: Tuple[PositiveFloat, PositiveFloat, PositiveFloat]
    crop_size: Tuple[PositiveInt, PositiveInt, PositiveInt]
    image_name: str
    temperature: float
    airy_unit: PositiveFloat
    bead_size: PositiveFloat
    bead_supplier: str
    mounting_medium: str
    operator: str
    microscope_type: str
    excitation: PositiveFloat
    emission: PositiveFloat
    comment: str
    dpi: PositiveInt



import pathlib
from copy import copy
from typing import Dict, Tuple

import numpy as np
from matplotlib import pyplot as plt
from matplotlib_scalebar.scalebar import ScaleBar
from mpl_toolkits.mplot3d.axis3d import Axis
from numpy._typing import ArrayLike
from pydantic import ValidationError

from psf_analysis_CFIM.psf_analysis.fit.fitter import YXFitter, ZFitter, ZYXFitter
from psf_analysis_CFIM.psf_analysis.image import Calibrated2DImage, Calibrated3DImage
from psf_analysis_CFIM.psf_analysis.records import (
    PSFRecord,
    YXFitRecord,
    ZFitRecord,
    ZYXFitRecord,
)
from psf_analysis_CFIM.psf_analysis.sample import YXSample

# Patch matplotlib to render ticks in 3D correctly.
# See: https://stackoverflow.com/a/16496436
if not hasattr(Axis, "_get_coord_info_old"):

    def _get_coord_info_new(self, renderer):
        mins, maxs, centers, deltas, tc, highs = self._get_coord_info_old(renderer)
        mins += deltas / 4
        maxs -= deltas / 4
        return mins, maxs, centers, deltas, tc, highs

    Axis._get_coord_info_old = Axis._get_coord_info
    Axis._get_coord_info = _get_coord_info_new


class PSFRenderEngine:
    from matplotlib.figure import Axes, Figure

    settings = {
        "covariance_ellipsoid": False,
        "coordinate_annotation": False,
    }
    psf_image: Calibrated3DImage = None
    psf_record: PSFRecord = None
    _figure: Figure = None
    _ax_xy: Axes = None
    _ax_yz: Axes = None
    _ax_zx: Axes = None
    _ax_3d: Axes = None
    _ax_3d_text: Axes = None

    def __init__(self, render_settings: dict, psf_image: Calibrated3DImage=None, psf_record: PSFRecord=None, ellipsoid_color="black", channels: dict = None):
        if render_settings:
            self.settings.update(render_settings)
        self.ellipsoid_color = ellipsoid_color
        if channels is not None:
            self.channels: dict[str:dict] = channels
            self.psf_image = psf_image
        else:
            self.psf_image = psf_image
            self.psf_record = psf_record
            self.ellipsoid_color = ellipsoid_color

    def _build_layout(self, dpi: int = 300) -> None:

        self._figure = plt.figure(figsize=(10, 10), dpi=dpi)
        self._add_axes()

    def _add_axes(self):
        self._ax_xy = self._figure.add_axes([0.025, 0.525, 0.45, 0.45])
        self._ax_yz = self._figure.add_axes([0.525, 0.525, 0.45, 0.45])
        self._ax_zx = self._figure.add_axes([0.025, 0.025, 0.45, 0.45])
        self._ax_3d = self._figure.add_axes(
            [0.55, 0.025, 0.42, 0.42], projection="3d", computed_zorder=False
        )

        self._ax_3d_text = self._figure.add_axes([0.525, 0.025, 0.45, 0.45])
        self._ax_3d_text.axis("off")

        self._add_axis_annotations()

    def _add_axis_annotations(self) -> None:
        ax_X = self._figure.add_axes([0.25, 0.5, 0.02, 0.02])
        ax_X.text(0, -0.1, "X", fontsize=14, ha="center", va="center")
        ax_X.axis("off")
        ax_Y = self._figure.add_axes([0.5, 0.75, 0.02, 0.02])
        ax_Y.text(0, 0, "Y", fontsize=14, ha="center", va="center")
        ax_Y.axis("off")
        ax_Z = self._figure.add_axes([0.5, 0.25, 0.02, 0.02])
        ax_Z.text(-0.5, 0, "Z", fontsize=14, ha="center", va="center")
        ax_Z.axis("off")
        ax_Z1 = self._figure.add_axes([0.75, 0.5, 0.02, 0.02])
        ax_Z1.text(0, 0.5, "Z", fontsize=14, ha="center", va="center")
        ax_Z1.axis("off")

    def render(
        self, date: str = None, version: str = None, dpi: int = 300, *, top_left_message: str = None, centroid: Tuple[float, float, float] = None
    ) -> ArrayLike:

        self._build_layout(dpi=dpi)
        self._add_projections(dpi=dpi)
        self._add_ellipsoids(centroid)

        if date is not None:
            self._add_date(date)

        if version is not None:
            self._add_version(version)

        if top_left_message:
            self._add_top_left_message(top_left_message)

        return self._fig_to_image()

    def _add_projections(self, dpi: int):
        self._add_yx_projection(dpi=dpi)
        self._add_zx_projection(dpi=dpi)
        self._add_yz_projection(dpi=dpi)

    def _add_yz_projection(self, dpi: int):
        yz_sqrt_projection = self._compute_centered_isotropic_image(
            projection=Calibrated2DImage(
                data=np.sqrt(np.max(self.psf_image.data, axis=2)).T,
                spacing=(self.psf_image.spacing[1], self.psf_image.spacing[0]),
            ),
            centroid=(self.psf_record.yx_fit.y_mu, self.psf_record.z_fit.z_mu),
            dpi=dpi,
        )
        self._add_image_to_axes(yz_sqrt_projection.data, self._ax_yz, origin="lower")
        self._draw_fwhm_annotations(
            self._ax_yz,
            shape=yz_sqrt_projection.data.shape,
            spacing=yz_sqrt_projection.spacing,
            fwhm_values=(self.psf_record.yx_fit.y_fwhm, self.psf_record.z_fit.z_fwhm),
            origin_upper=False,
        )

    def _add_zx_projection(self, dpi: int):
        zx_sqrt_projection = self._compute_centered_isotropic_image(
            projection=Calibrated2DImage(
                data=np.sqrt(np.max(self.psf_image.data, axis=1)),
                spacing=(self.psf_image.spacing[0], self.psf_image.spacing[2]),
            ),
            centroid=(self.psf_record.z_fit.z_mu, self.psf_record.yx_fit.x_mu),
            dpi=dpi,
        )
        self._add_image_to_axes(zx_sqrt_projection.data, self._ax_zx, origin="upper")
        self._draw_fwhm_annotations(
            self._ax_zx,
            shape=zx_sqrt_projection.data.shape,
            spacing=zx_sqrt_projection.spacing,
            fwhm_values=(
                self.psf_record.z_fit.z_fwhm,
                self.psf_record.yx_fit.x_fwhm,
            ),
            origin_upper=True,
        )

    def _add_yx_projection(self, dpi: int):
        yx_sqrt_projection = self._compute_centered_isotropic_image(
            projection=Calibrated2DImage(
                data=np.sqrt(np.max(self.psf_image.data, axis=0)),
                spacing=self.psf_image.spacing[1:],
            ),
            centroid=(self.psf_record.yx_fit.y_mu, self.psf_record.yx_fit.x_mu),
            dpi=dpi,
        )
        self._add_image_to_axes(yx_sqrt_projection.data, self._ax_xy, origin="lower")
        self._add_scalebar(yx_sqrt_projection.data.shape)
        self._draw_fwhm_annotations(
            self._ax_xy,
            shape=yx_sqrt_projection.data.shape,
            spacing=yx_sqrt_projection.spacing,
            fwhm_values=(
                self.psf_record.yx_fit.y_fwhm,
                self.psf_record.yx_fit.x_fwhm,
            ),
            origin_upper=False,
        )

    def _get_display_min_max(self, image: ArrayLike) -> Dict[str, float]:
        return {
            "vmin": np.quantile(image[image != -1], 0.03),
            "vmax": image[image != -1].max(),
        }

    def _add_image_to_axes(self, image: ArrayLike, ax: Axes, origin: str):
        ax.set_xticks([])
        ax.set_yticks([])
        ax.imshow(
            self._num_to_nan(image, num=-1),
            cmap=self._get_cmap(),
            interpolation="nearest",
            origin=origin,
            **self._get_display_min_max(image),
        )

    def _add_scalebar(self, shape: Tuple[int, int]):
        scalebar = ScaleBar(
            4000 / shape[1],
            "nm",
            fixed_value=500,
            location="lower right",
            border_pad=0.2,
            box_alpha=0.8,
        )
        self._ax_xy.add_artist(scalebar)

    def _get_cmap(self):
        from matplotlib import cm

        cmap = copy(cm.turbo)
        cmap.set_bad("white")
        return cmap

    def _num_to_nan(self, data: ArrayLike, num: float = -1):
        replaced = copy(data)
        replaced[replaced == num] = np.nan
        return replaced

    def _compute_centered_isotropic_image(
        self, projection: Calibrated2DImage, centroid: Tuple[float, float], dpi: int
    ) -> Calibrated2DImage:
        padded_projection = YXSample(
            image=Calibrated2DImage(
                data=np.pad(projection.data, 1, mode="constant", constant_values=(-1,)),
                spacing=projection.spacing,
            )
        )

        source_points = (
            padded_projection.get_ravelled_coordinates()
            - np.array([centroid])
            - np.array(padded_projection.image.spacing)
        )

        num_samples = self._estimate_number_of_samples(
            spacing=padded_projection.image.spacing,
            dpi=dpi,
        )
        target_points = self._get_target_points(num_samples=num_samples)

        import scipy

        return Calibrated2DImage(
            data=scipy.interpolate.griddata(
                points=source_points,
                values=padded_projection.image.data.ravel(),
                xi=target_points,
                method="nearest",
            ).reshape(num_samples, num_samples),
            spacing=(4000 / num_samples,) * 2,
        )

    def _get_target_points(self, num_samples: int):
        yy, xx = np.meshgrid(
            np.arange(num_samples) * 4000 / num_samples - 2000,
            np.arange(num_samples) * 4000 / num_samples - 2000,
            indexing="ij",
        )
        return np.stack([yy.ravel(), xx.ravel()], axis=-1)

    def _estimate_number_of_samples(self, spacing: Tuple[float, float], dpi: int):
        num_samples = int(np.min(spacing) * 10)
        while num_samples < 4.5 * dpi:
            num_samples += int(np.min(spacing) * 10)

        return num_samples

    def _draw_fwhm_annotations(
        self,
        axes: Axes,
        shape: Tuple[int, int],
        spacing: Tuple[float, float],
        fwhm_values: Tuple[float, float],
        origin_upper: bool,
    ) -> None:
        self._add_vertical_fwhm_annotation(
            axes=axes,
            fwhm_values=fwhm_values,
            shape=shape,
            spacing=spacing,
        )

        self._add_horizontal_fwhm_annotation(
            axes=axes,
            fwhm_values=fwhm_values,
            shape=shape,
            spacing=spacing,
            origin_upper=origin_upper,
        )

    def _add_horizontal_fwhm_annotation(
        self,
        axes: Axes,
        fwhm_values: Tuple[float, float],
        shape: Tuple[int, int],
        spacing: Tuple[float, float],
        origin_upper: bool,
    ):
        if origin_upper:
            shift_factor = 3
            start_factor = 1
            text_pos = shape[1] - shape[1] / 4.5
        else:
            shift_factor = 1
            start_factor = -1
            text_pos = shape[1] / 4.5

        cy = shape[0] / 2
        cx = shape[1] / 2
        y_fwhm, x_fwhm = fwhm_values

        if not np.isnan(y_fwhm) and not np.isnan(x_fwhm):
            dy = (y_fwhm / 2) / spacing[0]
            dx = (x_fwhm / 2) / spacing[1]

            axes.plot(
                [cx - dx, cx + dx],
                [
                    shift_factor * shape[1] / 4,
                ]
                * 2,
                linewidth=6,
                c="black",
                solid_capstyle="butt",
            )
            axes.plot(
                [cx - dx, cx + dx],
                [
                    shift_factor * shape[1] / 4,
                ]
                * 2,
                linewidth=4,
                c="white",
                solid_capstyle="butt",
            )
            axes.plot(
                [cx - dx, cx - dx],
                [cy + start_factor * dy, shift_factor * shape[1] / 4],
                "-",
                c="black",
            )
            axes.plot(
                [cx + dx, cx + dx],
                [cy + start_factor * dy, shift_factor * shape[1] / 4],
                "-",
                c="black",
            )
            axes.plot(
                [cx - dx, cx - dx],
                [cy + start_factor * dy, shift_factor * shape[1] / 4],
                "--",
                c="white",
            )
            axes.plot(
                [cx + dx, cx + dx],
                [cy + start_factor * dy, shift_factor * shape[1] / 4],
                "--",
                c="white",
            )

        axes.text(
            cx,
            text_pos,
            f"{self._get_fwhm_str(x_fwhm)}nm",
            ha="center",
            va="top",
            fontsize=15,
            bbox=dict(facecolor="white", alpha=0.8, linewidth=0),
        )

    def _get_fwhm_str(self, fwhm: float):
        if np.isnan(fwhm):
            return "NaN"
        else:
            return f"{int(np.round(fwhm))}"

    def _add_vertical_fwhm_annotation(
        self,
        axes: Axes,
        fwhm_values: Tuple[float, float],
        shape: Tuple[int, int],
        spacing: Tuple[float, float],
    ):
        cy = shape[0] / 2
        cx = shape[1] / 2
        y_fwhm, x_fwhm = fwhm_values

        if not np.isnan(y_fwhm) and not np.isnan(x_fwhm):
            dy = (y_fwhm / 2) / spacing[0]
            dx = (x_fwhm / 2) / spacing[1]

            axes.plot(
                [
                    shape[0] / 4,
                ]
                * 2,
                [cy - dy, cy + dy],
                linewidth=6,
                c="black",
                solid_capstyle="butt",
            )
            axes.plot(
                [
                    shape[0] / 4,
                ]
                * 2,
                [cy - dy, cy + dy],
                linewidth=4,
                c="white",
                solid_capstyle="butt",
            )
            axes.plot([shape[0] / 4, cx - dx], [cy - dy, cy - dy], "-", color="black")
            axes.plot([shape[0] / 4, cx - dx], [cy + dy, cy + dy], "-", color="black")
            axes.plot([shape[0] / 4, cx - dx], [cy - dy, cy - dy], "--", color="white")
            axes.plot([shape[0] / 4, cx - dx], [cy + dy, cy + dy], "--", color="white")

        axes.text(
            shape[1] / 4.5,
            cy,
            f"{self._get_fwhm_str(y_fwhm)}nm",
            ha="right",
            va="center",
            fontsize=15,
            bbox=dict(facecolor="white", alpha=0.8, linewidth=0),
        )

    def _has_nan_fwhm(self) -> bool:
        if np.isnan(self.psf_record.z_fit.z_fwhm):
            return True

        if np.isnan(self.psf_record.yx_fit.y_fwhm):
            return True

        if np.isnan(self.psf_record.yx_fit.x_fwhm):
            return True

        return False

    def _add_ellipsoids(self, centroid: Tuple[float, float, float]):
        self._configure_ticks_and_bounds()

        if not self._has_nan_fwhm():
            self._add_axis_aligned_ellipsoid() # Adds a grey ellipsoid from the fwhm of yx and z fits

        # TODO: Option in settings
        if self.settings["covariance_ellipsoid"]:
            self._add_cov_ellipsoid() # Adds a blue ellipsoid from covariance
        self._ax_3d_text.set_xlim(0, 100)
        self._ax_3d_text.set_ylim(0, 100)
        if centroid and self.settings["coordinate_annotation"]: # Disabled by request, TODO: Option in settings
            self._add_principal_components_annotons(centroid)

    def _add_color_ellipsoids(self, channel_offset_dict: dict = None):
        self._configure_ticks_and_bounds()

        self._add_fwhm_ellipsoids(channel_offset_dict)



    def _add_axis_aligned_ellipsoid(self):
        from psf_analysis_CFIM.psf_analysis.utils import sigma

        base_ell = self._get_simple_ellipsoid(
            xy_width=self.psf_record.yx_fit.y_fwhm, z_height=self.psf_record.z_fit.z_fwhm
        )
        self._ax_3d.plot_surface(
            *base_ell, rstride=2, cstride=2, color="grey", antialiased=True, alpha=1
        )
        self._ax_3d.plot_wireframe(
            *base_ell, rstride=3, cstride=3, color=self.ellipsoid_color, antialiased=True, alpha=0.5
        )

        self._ax_3d.contour(
            *base_ell,
            zdir="z",
            offset=self._ax_3d.get_zlim()[0],
            colors="white",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
        )
        self._ax_3d.contour(
            *base_ell,
            zdir="y",
            offset=self._ax_3d.get_ylim()[1],
            colors="white",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
        )
        self._ax_3d.contour(
            *base_ell,
            zdir="x",
            offset=self._ax_3d.get_xlim()[0],
            colors="white",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
        )
        self._ax_3d.contour(
            *base_ell,
            zdir="z",
            offset=self._ax_3d.get_zlim()[0],
            colors="black",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
            linestyles="--",
        )
        self._ax_3d.contour(
            *base_ell,
            zdir="y",
            offset=self._ax_3d.get_ylim()[1],
            colors="black",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
            linestyles="--",
        )
        self._ax_3d.contour(
            *base_ell,
            zdir="x",
            offset=self._ax_3d.get_xlim()[0],
            colors="black",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
            linestyles="--",
        )

    def _get_ellipsoid(
            self, covariance: ArrayLike, spacing: Tuple[float, float, float]
    ):
        bias = 0
        u = np.linspace(0, 2 * np.pi, 30)
        v = np.linspace(0, np.pi, 30)

        x = np.outer(np.cos(u), np.sin(v))
        y = np.outer(np.sin(u), np.sin(v))
        z = np.outer(np.ones_like(u), np.cos(v))
        cov_ = covariance / 2 * np.sqrt(2 * np.log(2))
        x, y, z = (cov_ @ np.stack((x, y, z), 0).reshape(3, -1) + bias).reshape(
            3, *x.shape
        )
        return x / spacing[2], y / spacing[1], z / spacing[0]

    def _get_simple_ellipsoid(self, xy_width: float | tuple[float, float], z_height: float, resolution: int = 30, zyx_offset: Tuple[float, float, float] = (0, 0, 0)):
        if isinstance(xy_width, float):
            a = xy_width / 2
            b = xy_width / 2
        else:
            a = xy_width[0] / 2
            b = xy_width[1] / 2


        c = z_height / 2

        # Create angular grids.
        u = np.linspace(0, 2 * np.pi, resolution)
        v = np.linspace(0, np.pi, resolution)

        # Parametric equations for the ellipsoid.
        x = a * np.outer(np.cos(u), np.sin(v))
        y = b * np.outer(np.sin(u), np.sin(v))
        z = c * np.outer(np.ones_like(u), np.cos(v))

        return x + zyx_offset[2], y + zyx_offset[1], z + zyx_offset[0]

    def _add_fwhm_ellipsoids(self, channel_offset_dict):

        import math


        # Sort channels based on the distance in (z, -y, x) space.
        sorted_keys = sorted(
            self.channels,
            key=lambda key: math.sqrt(
                (channel_offset_dict.get(key, (0, 0, 0))[0]) ** 2 +
                (-channel_offset_dict.get(key, (0, 0, 0))[1]) ** 2 +
                (channel_offset_dict.get(key, (0, 0, 0))[2]) ** 2
            ) if channel_offset_dict else 0,
        )

        channel_dim_prio_dict = {}
        dims = {"z": 0, "y": 1, "x": 2}
        for axis, idx in dims.items():
            sorted_keys = sorted(channel_offset_dict.keys(), key=lambda k: channel_offset_dict[k][idx], reverse=False)
            channel_dim_prio_dict[axis] = tuple(sorted_keys)

        ellipsoid_calc_dict = {}
        for key in sorted_keys:
            channel_dict = self.channels[key]
            zyx_fwhm = channel_dict.get("zyx")
            color = channel_dict["color"]
            if channel_offset_dict:
                channel_offset = channel_offset_dict.get(key, (0, 0, 0))
            else:
                channel_offset = (0, 0, 0)

            if channel_offset == (0, 0, 0):
                print(f"Found no offset for channel {key}")


            ellipsoid_calc = self._get_simple_ellipsoid(
                xy_width=(zyx_fwhm[2], zyx_fwhm[1]), z_height=zyx_fwhm[0], zyx_offset=channel_offset
            )

            self._ax_3d.plot_surface(
                *ellipsoid_calc, rstride=1, cstride=1, color=color, antialiased=True, alpha=0.3
            )
            self._ax_3d.plot_wireframe(
                *ellipsoid_calc, rstride=4, cstride=4, color=color, antialiased=True, alpha=0.2
            )

            ellipsoid_calc_dict[key] = ellipsoid_calc

        # We use the dim sorted dict to paint the ellipsoids in the correct order.
        # Kinda ugly, but thats what I could think of.
        zyx_args_dict = {
            "z": (self._ax_3d.get_zlim()[0], 0),
            "y": (self._ax_3d.get_ylim()[1], 1),
            "x": (self._ax_3d.get_xlim()[0], 2),
        }

        for dim_key in channel_dim_prio_dict:

            for channel_key in channel_dim_prio_dict[dim_key]:
                ellipsoid_calc = ellipsoid_calc_dict[channel_key]
                color = self.channels[channel_key]["color"]
                zyx_offset = channel_offset_dict.get(channel_key, (0, 0, 0))

                lim = zyx_args_dict[dim_key][0]
                dim_offset = zyx_offset[zyx_args_dict[dim_key][1]]


                self._ax_3d.contour(
                    *ellipsoid_calc,
                    zdir=dim_key,
                    offset=lim,
                    levels=[dim_offset],
                    colors=color,
                    linestyles="solid",
                )

                self._ax_3d.contourf(
                        *ellipsoid_calc,
                        zdir=dim_key,
                        offset=lim,
                        colors=color,
                        alpha=0.1,
                        levels=[dim_offset, dim_offset + 1000],
                    )


    def _add_cov_ellipsoid(self):
        fit = self.psf_record.zyx_fit
        covariance = np.array(
            [
                [fit.zyx_cxx, fit.zyx_cyx, fit.zyx_czx],
                [fit.zyx_cyx, fit.zyx_cyy, fit.zyx_czy],
                [fit.zyx_czx, fit.zyx_czy, fit.zyx_czz],
            ]
        )

        cv_ell = self._get_ellipsoid(
            covariance=covariance, spacing=self.psf_image.spacing
        )
        self._ax_3d.plot_surface(
            *cv_ell, rstride=1, cstride=1, color="navy", antialiased=True, alpha=0.15
        )
        self._ax_3d.plot_wireframe(
            *cv_ell, rstride=4, cstride=4, color="navy", antialiased=True, alpha=0.25
        )


        self._ax_3d.contour(
            *cv_ell,
            zdir="z",
            offset=self._ax_3d.get_zlim()[0],
            colors="navy",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
        )
        self._ax_3d.contour(
            *cv_ell,
            zdir="y",
            offset=self._ax_3d.get_ylim()[1],
            colors="navy",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
        )
        self._ax_3d.contour(
            *cv_ell,
            zdir="x",
            offset=self._ax_3d.get_xlim()[0],
            colors="navy",
            levels=1,
            vmin=-1,
            vmax=1,
            zorder=0,
        )

    def _configure_ticks_and_bounds(self, aspect_override: Tuple[float, float, float] = None):
        labels_list = [-1000, -750, -500, -250, 0, 250, 500, 750, 1000]
        if aspect_override:
            self._ax_3d.set_box_aspect(aspect_override)
        else:
            self._ax_3d.set_box_aspect((1.0, 1.0, 4.0 / 3.0))
        self._ax_3d.set_xticks(labels_list)
        self._ax_3d.set_xticklabels(
            labels_list, verticalalignment="bottom", horizontalalignment="right"
        )
        self._ax_3d.set_yticks(labels_list)
        self._ax_3d.set_yticklabels(
            labels_list, verticalalignment="bottom", horizontalalignment="left"
        )

        self._ax_3d.set_zticks(labels_list)
        self._ax_3d.set_zticklabels(
            labels_list,
            verticalalignment="top",
            horizontalalignment="left",
        )
        self._ax_3d.set_xlim(-800, 800)
        self._ax_3d.set_ylim(-800, 800)
        self._ax_3d.set_zlim(-1100, 1100)

    def _add_principal_components_annotons(self,centroid):

        self._ax_3d_text.text(
            -5,
            95,
            "Coordinates (nm)",
            fontsize=14,
            weight="bold",
            color="navy",
        )
        self._ax_3d_text.text(
            0,
            89,
            f"Z|{round(centroid[0],1)}",
            fontsize=14,
            color="navy",
        )
        self._ax_3d_text.text(
            0,
            83,
            f"Y|{round(centroid[1],1)}",
            fontsize=14,
            color="navy",
        )
        self._ax_3d_text.text(
            0,
            77,
            f"X|{round(centroid[2],1)}",
            fontsize=14,
            color="navy",
        )

    def _add_date(self, date: str) -> None:
        self._ax_3d_text.text(
            100,
            -4,
            f"Acquisition date: {date}",
            fontsize=11,
            horizontalalignment="right",
        )

    def _add_version(self, version: str) -> None: # Adds text to the bottom left with version number
        self._ax_3d_text.text(-110, -4, f"psf-analysis-CFIM: v{version}", fontsize=11)

    def _add_top_left_message(self, message: str) -> None:
        self._ax_3d_text.text(-110, 212, message, fontsize=14)

    def _fig_to_image(self):
        from matplotlib_inline.backend_inline import FigureCanvas

        canvas = FigureCanvas(self._figure)
        canvas.draw()
        image = np.frombuffer(canvas.tostring_rgb(), dtype="uint8").reshape(
            (self._figure.canvas.get_width_height()[::-1]) + (3,)
        )

        plt.close(self._figure)
        return image

    def render_multi_channel_summary(self, channel_offset_dict: dict, table_beads:dict,
                                     date: str = None, version: str = None, microscope: str = None, objective: str = None):
        """
            Renders a summary image with an ellipsoid for each color channel.
        """
        dpi = 300
        self._figure = plt.figure(figsize=(10, 10), dpi=dpi / 2) # TODO:  use dpi from input, might need to be dpi / 2
        self._ax_3d = self._figure.add_subplot(111, projection="3d")
        pos = self._ax_3d.get_position()
        # Apparently a cm is about 0.0398 inches | And the figure is in inches r/todayIlearned
        self._ax_3d.set_position([pos.x0, pos.y0 - 0.0787, pos.width, pos.height])
        table_ax = self._figure.add_axes((0.06, 0.790, 0.3, 0.16))


        # Allocate an additional axes for the distance table.
        self.create_distance_table_ax(table_beads, table_ax)

        self._add_color_ellipsoids(channel_offset_dict=channel_offset_dict)

        # Add the logo in the corner
        l_path = pathlib.Path(__file__).parent.parent / "resources" / "logo.png"
        logo = plt.imread(l_path)
        ax_logo = self._figure.add_axes([0.79, 0.82, 0.2, 0.2], zorder=2)
        ax_logo.imshow(logo)

        if date is not None:
            self._figure.text(0.99, 0.01, f"Acquisition date: {date}", ha="right", va="bottom", fontsize=12)


        # Adding text to the bottom left of the figure with two extra lines above
        if microscope:  self._figure.text(0.01, 0.059, f"Microscope: {microscope}", ha="left", va="bottom", fontsize=12)
        if objective:   self._figure.text(0.01, 0.04, f"Objective: {objective}", ha="left", va="bottom", fontsize=12)
        if version:     self._figure.text(0.01, 0.01, f"psf-analysis-CFIM: v{version}", ha="left", va="bottom", fontsize=12)



        return self._fig_to_image()

    def create_distance_table_ax(self, beads, ax):
        """
        Creates a table showing pairwise Euclidean distances between beads on an existing axes.
        Each bead is provided as a dictionary with:
           - 'bead': a tuple/list of zyx coordinates.
           - 'colormap': a string with a valid matplotlib color.
        """
        # Sort keys numerically for consistency.
        bead_keys_sorted = sorted(beads.keys(), key=lambda x: float(x))
        keys = [f"{wavelength}λ" for wavelength in bead_keys_sorted]
        n = len(keys)

        # Compute pairwise distance matrix.
        dist_matrix = np.zeros((n, n))
        for i, key_i in enumerate(bead_keys_sorted):
            coord_i = np.array(beads[key_i]["bead"])
            for j, key_j in enumerate(bead_keys_sorted):
                coord_j = np.array(beads[key_j]["bead"])
                dist_matrix[i, j] = np.linalg.norm(coord_i - coord_j)

        # Format the values for table cells.
        cell_text = [[f"{dist_matrix[i, j]:.0f}" for j in range(n)] for i in range(n)]

        # Create the table on the provided axes.
        table = ax.table(cellText=cell_text, rowLabels=keys, colLabels=keys, loc='upper center')
        table.scale(1, 1.5)

        # Color according to colormap
        cells = table.get_celld()
        for i, key in enumerate(bead_keys_sorted):
            row_cell = cells.get((0, i))
            col_cell = cells.get((i + 1, -1))
            for cell in (row_cell, col_cell):
                if cell:
                    cell.set_facecolor(beads[key]["colormap"])
                    cell.get_text().set_color("white")

        title = ax.set_title("Distances between centers in nm", color="black", fontsize=14)
        title.set_position([0.45, 1])

        ax.axis('tight')
        ax.axis('off')



class PSF:
    image: Calibrated3DImage = None
    psf_record: PSFRecord = None
    error = False
    settings: dict = {}

    def __init__(self, image: Calibrated3DImage, psf_settings: dict):
        self.image = image
        self.settings = psf_settings

    def analyze(self) -> None:
        z_fitter = ZFitter(image=self.image)
        yx_fitter = YXFitter(image=self.image)
        zyx_fitter = ZYXFitter(image=self.image)
        try:
            z_fit_record: ZFitRecord = z_fitter.fit()
            yx_fit_record: YXFitRecord = yx_fitter.fit()
            zyx_fit_record: ZYXFitRecord = zyx_fitter.fit()
        except RuntimeError as e:
            print(f"Runtime error fitting point: {self.image.get_corner_coordinates()}")
            self.error = True
        except (TypeError, ValidationError) as e:
            print(f"Type error fitting point: {self.image.get_corner_coordinates()} | error: {e}")
            self.error = True
        else:
            self.psf_record = PSFRecord(
                z_fit=z_fit_record,
                yx_fit=yx_fit_record,
                zyx_fit=zyx_fit_record,
            )


    def get_record(self) -> PSFRecord:
        return self.psf_record

    def get_summary_image(
        self,
        date: str = None,
        version: str = None,
        dpi: int = 300,
        *,
        top_left_message: str = None,
        ellipsoid_color: str = "black",
        centroid = None,
    ) -> ArrayLike:
        engine = PSFRenderEngine(render_settings=self.settings.get("render_settings"),psf_image=self.image, psf_record=self.psf_record, ellipsoid_color=ellipsoid_color)
        return engine.render(date=date, version=version, dpi=dpi, top_left_message=top_left_message, centroid=centroid)

    def get_summary_dict(self) -> dict:
        return {
            **self.psf_record.z_fit.dict(),
            **self.psf_record.yx_fit.dict(),
            **self.psf_record.zyx_fit.dict(),
        }

    def get_fwhm(self):
        return self.psf_record.z_fit.z_fwhm,  self.psf_record.yx_fit.y_fwhm, self.psf_record.yx_fit.x_fwhm




class RangeDict:
    def __init__(self, range_map: list[tuple]):
        self.range_map = range_map

    def __getitem__(self, key):
        for lower, upper, value in self.range_map:
            if lower <= key < upper:
                return value
        raise KeyError(f"{key} not found in any range")


from typing import Tuple, List

import numpy as np
from scipy.ndimage import median_filter
from skimage.feature import peak_local_max


class BeadFinder:
    def __init__(self, image_layers, scale: tuple, bounding_box: tuple | list[tuple], bead_finder_settings: dict):

        self.settings = bead_finder_settings
        self._debug = bead_finder_settings.get("debug", False)

        if isinstance(bounding_box, list):
            max_zyx = (0, 0, 0)
            for box in bounding_box:
                if not isinstance(box, tuple):
                    raise ValueError("Bounding box must be a tuple or list of tuples.")
                if len(box) != 3:
                    raise ValueError("Bounding box must be a tuple of 3 integers.")
                max_zyx = tuple(max(max_zyx[i], box[i]) for i in range(3))
            bounding_box = max_zyx
        if self._debug:
            print(f"Bounding box set to {bounding_box}")

        self.bounding_box_px = np.array(bounding_box) / np.array(scale)
        self.max_bead_dist = np.linalg.norm(np.array(self.bounding_box_px)) / 2

        self.images_list = image_layers
        self.scale = scale

        self.yx_border_padding = 2 # TODO: Add to settings
        self.z_border_padding = 0

        self.maxima_rel = bead_finder_settings.get("maxima_rel", 0.2)
        self.maxima_abs = bead_finder_settings.get("maxima_abs", 0)

        self.passed_bead_count = [0, 0, 0]
        self.discarded_bead_count = [0, 0, 0]


    """
       A class for finding approximate bead positions in a 3D image stack.
       
        Parameters:
        - images_list: A list of napari image layers. 
            Multiple layers are assumed to be different channels / wavelengths.
        - scale: A tuple of 3 floats representing the pixel size for display in napari.
        - bounding_box: tuple[int, int, int] -> represent the psf box in nm.
        
        Returns:
        - A list of dicts, 
            points: list[tuple(z, y, x) for found beads, 
            discarded: list[tuple(z, y, x) for discarded beads,
            wavelength: an int, for identifying the channel.
    """

    def find_beads(self):
        total_beads = []
        total_discarded_beads = []
        channels_beads_dicts = []


        for image_layer in self.images_list:
            try:
                wavelength = image_layer.metadata["EmissionWavelength"]
            except KeyError:
                wavelength = None

            beads, discarded_beads = self.find_beads_for_channel(image_layer.data)

            total_beads += beads
            total_discarded_beads += discarded_beads

            channel_beads_dict = {
                "points": beads,
                "discarded": discarded_beads,
                "wavelength": wavelength
            }

            channels_beads_dicts.append(channel_beads_dict)

        # if self._debug: # It was really important to color code the output, trust me... # I should color each channel separately :D
        #     green = '\033[92m'
        #     yellow = '\033[93m'
        #     endc = '\033[0m'
        #     passed_discarded_beads = len(total_discarded_beads) + len(total_beads)
        #     # TODO: Have this count correctly for multiple channels
        #     print(
        #         f"Beads {green}passed{endc} / {yellow}discarded{endc} of {passed_discarded_beads} for {len(self.images_list)} color channel{"s" if len(self.images_list) > 1 else ""} \nxy border: {green}{self.passed_bead_count[0]}{endc} / {yellow}{self.discarded_bead_count[0]}{endc} "
        #         f"| z border: {green}{self.passed_bead_count[1]}{endc} / {yellow}{self.discarded_bead_count[1]}{endc} | neighbor dist: {green}{self.passed_bead_count[2]}{endc} / {yellow}{self.discarded_bead_count[2]}{endc}")
        #     print(f"Total: {green}{len(total_beads)}{endc} / {yellow}{len(total_discarded_beads)}{endc}")
        return channels_beads_dicts

    def find_beads_for_channel(self, channel_image):
        median_image = self._median_filter(self._max_projection(channel_image))

        yx_beads, discarded_xy = self._maxima(median_image, channel_image)
        zyx_beads, zyx_discarded_beads = self._find_bead_positions(channel_image, yx_beads)

        beads, discarded_beads_by_neighbor_dist = self.filter_beads_by_neighbour_distance(zyx_beads, zyx_discarded_beads)
        yx_discarded_beads, x = self._find_bead_positions(channel_image, discarded_xy, no_filter=True) # Convert discarded yx beads to zyx

        # Combine discarded beads TODO: Add lines to visualize discarded beads from neighbor distance

        if self._debug:
            self.passed_bead_count[0] += len(yx_beads)
            self.passed_bead_count[1] += len(zyx_beads)
            self.passed_bead_count[2] += len(beads)
            self.discarded_bead_count[0] += len(discarded_xy)
            self.discarded_bead_count[1] += len(zyx_discarded_beads)
            self.discarded_bead_count[2] += len(discarded_beads_by_neighbor_dist)

        discarded_beads = zyx_discarded_beads + yx_discarded_beads + discarded_beads_by_neighbor_dist
        return beads, discarded_beads

    def get_scale(self):
        return self.scale

    def _max_projection(self, image=None):
        return np.max(image, axis=0)

    def _median_filter(self, image, size=3):
        return median_filter(image, size=size)

    # TODO: Make threshold a setting with option for rel/abs
    def _maxima(self, image, channel_image) -> (List[Tuple], List[Tuple]):

        yx_border = (self.bounding_box_px[1] / 2) + self.yx_border_padding
        image_size = channel_image.shape
        xy_bead_positions = peak_local_max(image, min_distance=2, threshold_rel=self.maxima_rel, threshold_abs=self.maxima_abs, exclude_border=0)
        xy_bead_positions = [(y, x) for (y, x) in xy_bead_positions]
        in_border_xy_bead_positions = [bead for bead in xy_bead_positions if yx_border < bead[0] < image_size[1] - yx_border and yx_border < bead[1] < image_size[2] - yx_border]
        discarded_beads = [bead for bead in xy_bead_positions if bead not in in_border_xy_bead_positions]

        return in_border_xy_bead_positions, discarded_beads

    def _find_bead_positions(self, image, xy_beads, no_filter=False):
        bead_pos = []
        discarded_beads = []
        z_border = self.bounding_box_px[0] / 2
        for (y, x) in xy_beads:
            z_profile = image[:, y, x]

            z_profile_median = self._median_filter(z_profile, size=2)

            z = np.argmax(z_profile_median)
            if 0 + z_border < z < image.shape[0] - z_border or no_filter:
                bead_pos.append((z, y, x))
            else:
                discarded_beads.append((z, y, x))

        return bead_pos, discarded_beads

    def filter_beads_by_neighbour_distance(self, beads, discarded_beads):
        discarded_beads = []
        valid_beads = []
        half_box = self.bounding_box_px / 2.0

        all_beads = beads + discarded_beads
        for bead in beads:
            is_valid = True
            for neighbour in all_beads:
                if bead == neighbour:
                    continue
                # Check if each coordinate of neighbour is within bead \+\/\- half_box
                if (bead[0] - half_box[0] <= neighbour[0] <= bead[0] + half_box[0] and
                        bead[1] - half_box[1] <= neighbour[1] <= bead[1] + half_box[1] and
                        bead[2] - half_box[2] <= neighbour[2] <= bead[2] + half_box[2]):
                    is_valid = False
                    break
            if is_valid:
                valid_beads.append(bead)
            else:
                discarded_beads.append(bead)
        return valid_beads, discarded_beads


    def set_settings(self, settings_dict):
        self.yx_border_padding = settings_dict['yx_border_padding']
        self.z_border_padding = settings_dict['z_border_padding']
        self.maxima_rel = settings_dict['maxima_relative']
        self.maxima_abs = settings_dict['maxima_absolute']
        self.max_bead_dist = settings_dict['max_bead_distance']



from typing import Tuple

import numpy as np
from numpy._typing import ArrayLike
from skimage.measure import centroid


def sigma(fwhm: float) -> float:
    """Compute sigma from full width half maximum.

    Parameters
    ----------
    fwhm :
        full width half maximum

    Returns
    -------
    float
        sigma
    """
    return fwhm / (2 * np.sqrt(2 * np.log(2)))


def fwhm(sigma: float) -> float:
    """Full width at half maximum.

    Compute full width at half maximum (FWHM) for the provided standard
    deviation of `sigma`.

    Parameters
    ----------
    sigma :
        standard deviation

    Returns
    -------
    float
        full width at half maximum
    """
    return 2 * np.sqrt(2 * np.log(2)) * sigma


def compute_cov_matrix(img_data: ArrayLike, spacing: Tuple[float, ...]) -> ArrayLike:
    """Weighted covariance matrix.

    Given some image data the intensity weighted covariance matrix over the
    spatial dimensions is computed.

    Parameters
    ----------
    img_data :
        n-dimensional image
    spacing :
        pixel spacing

    Returns
    -------
        The covariance matrix of the image data weighted by the intensity.
    """
    extends = [np.arange(dim_size) * s for dim_size, s in zip(img_data.shape, spacing)]

    grids = np.meshgrid(*extends, indexing="ij")

    m = np.stack([g.ravel() for g in grids])

    return np.cov(m, fweights=img_data.ravel())


def estimate_from_data(
    sample: ArrayLike, data: ArrayLike, sample_spacing: Tuple[float, ...]
):
    """Estimate Gaussian fit parameters from data.

    Given some sample and data initialization parameters for a Gaussian fit
    are estimated. The following parameters are computed:
    * background: Is the median over all `data`.
    * Amplitude: Maximum of `sample`.
    * mean: Weighted centroid of the background subtracted sample.
    * sigma: Standard deviation estimated from the covariance matrix.

    Parameters
    ----------
    sample :
        A subset of data for which the estimates are computed.
    data :
        The whole data, used to get a better estimate for the background.
    sample_spacing :
        Pixel spacing of `sample`.

    Returns
    -------
    background:
        Estimated data background
    amplitude:
        Estimated Gaussian amplitude
    mu:
        Weighed centroid of the sample
    sigma:
        Standard deviation of the sample
    """
    bg = np.median(data).astype(np.uint16)
    amp = sample.max() - bg

    z_sample_no_bg = np.clip(sample - bg, 0, data.max())

    mu = centroid(z_sample_no_bg) * np.array(sample_spacing)
    cov_matrix = compute_cov_matrix(img_data=z_sample_no_bg, spacing=sample_spacing)
    if cov_matrix.ndim == 0:
        sigma = np.sqrt(cov_matrix)
    else:
        sigma = np.sqrt(np.diag(cov_matrix))
    return bg, amp, mu, sigma



from pydantic import BaseModel, Field, conint, confloat, conlist
from typing import List
import os

def get_default_output_folder() -> str:
    base = os.getenv('LOCALAPPDATA', os.path.expanduser('~\\AppData\\Local'))
    return os.path.join(base, "psf-analysis-cfim", "output")

class BeadFinderSettings(BaseModel):
    debug: bool = False
    maxima_rel: float = 0.2
    maxima_abs: float = 0

class RenderSettings(BaseModel):
    covariance_ellipsoid: bool = False
    coordinate_annotation: bool = False

class PSFSettings(BaseModel):
    render_settings: RenderSettings = RenderSettings()

class AnalyzerSettings(BaseModel):
    psf_settings: PSFSettings = PSFSettings()
    debug: bool = False

class IntensitySettings(BaseModel):
    lower_warning_percent: float = Field(0.08, ge=0, le=1)
    lower_error_percent: float = Field(0.12, ge=0, le=1)
    upper_warning_percent: float = Field(0.01, ge=0, le=1)
    upper_error_percent: float = Field(0.08, ge=0, le=1)


class NoiseSettings(BaseModel):
    high_noise_threshold: conint(ge=0) = 120
    low_snr_threshold: conint(ge=0) = 10

class ImageAnalysisSettings(BaseModel):
    intensity_settings: IntensitySettings = IntensitySettings()
    noise_settings: NoiseSettings = NoiseSettings()


class UISettings(BaseModel):
    bead_size: conint(gt=0) = 100
    box_size_xy: conint(gt=0) = 2000
    box_size_z: conint(gt=0) = 2500
    ri_mounting_medium: confloat(gt=0.9, lt=2) = 1.4

class PSFAnalysisPluginSettings(BaseModel):
    __version__: str = "1.7.10"

    debug: bool = Field(default=False)
    version: str = Field(default=__version__)
    microscopes: List[str] = ["TIRF", "Zeiss Z1"]
    output_folder: str = Field(default_factory=get_default_output_folder)
    ui_settings: UISettings = UISettings()
    image_analysis_settings: ImageAnalysisSettings = ImageAnalysisSettings()
    analyzer_settings: AnalyzerSettings = AnalyzerSettings()
    bead_finder_settings: BeadFinderSettings = BeadFinderSettings()




import os

from reportlab.lib import colors
from qtpy.QtWidgets import QGroupBox, QFormLayout, QWidget, QHBoxLayout, QFileDialog, QLineEdit, QPushButton, QLabel
from reportlab.lib.pagesizes import letter
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.units import inch
from reportlab.platypus import SimpleDocTemplate, Table, TableStyle, Paragraph, Spacer


class WarningMessage:
    def __init__(self, message:str, is_ignored:bool=False):
        self.message = message
        self.is_ignored = is_ignored

class WarningMessageList:
    def __init__(self, list_of_warnings: list[WarningMessage]=[]):
        self.warnings = list_of_warnings

    def add_warning(self, warning):
        self.warnings.append(warning)

class ErrorMessageList(WarningMessageList):
    def __init__(self, list_of_warnings: list[WarningMessage]=[]):
        super().__init__(list_of_warnings)

    def add_error(self, error):
        self.warnings.append(error)


def _get_unique_filepath(path):
    """
        If the filepath already exists, appends a counter to the base name.
        E.g., if 'example_report.pdf' exists, it returns 'example_report(1).pdf'.
        """
    if not os.path.exists(path):
        return path

    base, ext = os.path.splitext(path)
    counter = 1
    new_filepath = f"{base}({counter}){ext}"
    while os.path.exists(new_filepath):
        counter += 1
        new_filepath = f"{base}({counter}){ext}"
    return new_filepath





class ReportWidget(QWidget):
    def __init__(self,parent=None, filename = "Report", title = "Report", output_path=None):
        super().__init__(parent)
        self.border_color = colors.green
        self.bead_variation = None
        self.filename = filename
        self.title = title
        self.bead_data: dict = {}
        self._canvas = None
        self._path = None
        self._output_path = output_path
        self._bead_keys = ["z_fwhm", "y_fwhm", "x_fwhm"]
        self.save_dir_line_edit = None
        self.save_path = None

        # Don't ask why I made these into their own types, I just did.
        self.warnings = WarningMessageList()
        self.errors = ErrorMessageList()

    def set_title(self, title):
        self.title = title

    def add_bead_stats(self, stats: dict, data_type="Average bead", channel="None"):
        """
            Add bead stats to the report.
        """
        bead_stats = {f"{data_type}": stats}

        if channel in self.bead_data:
            self.bead_data[channel].update(bead_stats)
        else:
            self.bead_data[channel] = bead_stats



    def add_bead_stats_psf(self, stats, title="Average bead"):
        """
            Reformats a PSFRecord and calls add_bead_stats.
        """
        bead_stats = {
            "z_fwhm": stats.z_fit.z_fwhm,
            "y_fwhm": stats.yx_fit.y_fwhm,
            "x_fwhm": stats.yx_fit.x_fwhm,
        }
        self.add_bead_stats(bead_stats, title)
    def set_averaged_bead(self, bead):
        """
            Set the averaged bead.
        """
        self.bead_variation = bead

    def set_bead_variation(self, stats, stats_type="Bead Variation", channel=None):
        """
            Set the bead variation.
        """
        variation_bead = {
            "stats_type": stats_type,
            "channel": channel,
            "z_fwhm": f"{int(stats['z_fwhm_max'])} - {int(stats['z_fwhm_min'])}",
            "y_fwhm": f"{int(stats['y_fwhm_max'])} - {int(stats['y_fwhm_min'])}",
            "x_fwhm": f"{int(stats['x_fwhm_max'])} - {int(stats['x_fwhm_min'])}",
        }
        self.bead_variation = variation_bead

    def add_warnings_and_errors(self, warnings:WarningMessageList, errors:ErrorMessageList):
        """
            Add warnings and errors to the report.
        """
        self.warnings.add_warning(warnings)
        self.errors.add_error(errors)

    def create_pdf(self, path=None):

        self._path = path

        if not os.path.exists(self._path):
            raise FileNotFoundError(f"Directory {self._path} does not exist.")

        path_filename = os.path.join(self._path, self.filename + ".pdf")

        filepath = _get_unique_filepath(path_filename)
        #
        # c = canvas.Canvas(filename=unique_filepath, pagesize=letter, lang="en-gb")
        # c.setTitle(self.title)
        #
        # width, height = letter
        #
        # # Set title font and size
        # c.setFont("Helvetica-Bold", 24)
        # # Draw the title centered on the page
        # c.drawCentredString(width / 2.0, height - 100, self.title)
        #
        # # Additional content like stats, warnings, errors, and grade can be added here.
        # # For example, to add some text:
        # c.setFont("Helvetica", 12)
        # c.drawString(100, height - 150, "Overall stats, warnings, and grade will appear here.")
        #
        # c.showPage()
        # self._canvas = c
        # Create a PDF document using Platypus


        doc = SimpleDocTemplate(filepath, pagesize=letter)
        doc.title = self.title
        styles = getSampleStyleSheet()
        story = []

        # Add the title
        story.append(Paragraph(self.title, styles["Title"]))
        story.append(Spacer(1, 12))

        # Add the bead variation if it exists
        if self.bead_variation:
            data = self.bead_data + [self.bead_variation]
        else:
            data = self.bead_data

        # Create table data with header and measurement rows.
        table_data = self.beads_to_schema(data, styles["BodyText"])

        # Create the table and set a style for it
        table = Table(table_data, hAlign="LEFT")
        table.setStyle(TableStyle([
            ('BACKGROUND', (0, 0), (-1, 0), colors.lightgrey),
            ('TEXTCOLOR', (0, 0), (-1, 0), colors.black),
            ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
            ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
            ('BOTTOMPADDING', (0, 0), (-1, 0), 12),
            ('GRID', (0, 0), (-1, -1), 1, colors.black),
        ]))

        story.append(table)
        doc.build(
            story,
            onFirstPage=lambda canvas, doc: self.draw_border(canvas, doc),
            onLaterPages=lambda canvas, doc: self.draw_border(canvas, doc)
        )
        print(f"Report saved as: {filepath}")

    def draw_border(self,canvas, doc):
        """
            Draws a border on the current canvas with the specified border_color.
            """
        canvas.saveState()
        canvas.setStrokeColor(self.border_color)
        canvas.setLineWidth(4)  # Adjust the thickness as needed
        margin = 0.5 * inch  # Set margin from the page edge
        width, height = doc.pagesize
        canvas.rect(margin, margin, width - 2 * margin, height - 2 * margin)
        canvas.restoreState()

    def init_ui(self):
        pane = QGroupBox(parent=self)
        pane.setLayout(QFormLayout())

        dir_selection_dialog = QWidget(parent=self)
        dir_selection_dialog.setLayout(QHBoxLayout())

        self.save_path = QFileDialog()
        self.save_path.setFileMode(QFileDialog.DirectoryOnly)
        self.save_path.setDirectory(
            str(self._output_path)
        )

        self.save_dir_line_edit = QLineEdit()
        self.save_dir_line_edit.setText(self.save_path.directory().path())

        choose_dir = QPushButton("...")
        choose_dir.clicked.connect(self.parent().select_save_dir)

        dir_selection_dialog.layout().addWidget(
            QLabel("Save Dir", dir_selection_dialog)
        )
        dir_selection_dialog.layout().addWidget(self.save_dir_line_edit)
        dir_selection_dialog.layout().addWidget(choose_dir)

        pane.layout().addRow(dir_selection_dialog)

        self.save_button = QPushButton("Save Measurements")
        self.save_button.setEnabled(True)
        self.save_button.clicked.connect(self.parent().save_measurements)

        pane.layout().addRow(self.save_button)

        return pane

    def beads_to_schema(self, beads, style):
        """
        Convert a list of bead measurement dictionaries to a table schema.

        """
        # Define header
        headers = [Paragraph(item, style) for item in ["Bead"] + self._bead_keys]
        schema = [headers]

        # Append each bead's measurements in order
        for bead in beads:
            row = [
                Paragraph(str(bead.get("title", "N/A")), style),
                Paragraph(str(bead.get("z_fwhm", "N/A")), style),
                Paragraph(str(bead.get("y_fwhm", "N/A")), style),
                Paragraph(str(bead.get("x_fwhm", "N/A")), style)
            ]
            schema.append(row)
        return schema

# if __name__ == "__main__":
#     path_string = "C:/napari-psf-analysis-CFIM-edition/src/psf_analysis_CFIM/report_widget/dev_output"
#     report = ReportWidget(filename="example_report.pdf", "PSF Analysis Report")
#     report.create_pdf(path=path_string)
#     report.save_pdf(path_string)


import os

from napari.utils.notifications import show_warning
import importlib.util

from psf_analysis_CFIM.czi_reader.czi_metadata_processor import extract_key_metadata
import numpy as np

# Fail fast at import time if `bfio` is present; `bfio` is known to
# cause circular imports inside the optional `aicsimageio` bfio reader
# which results in confusing import-time tracebacks. Raising a clear
# error here gives the user an actionable message immediately.
BFIO_PRESENT = importlib.util.find_spec("bfio") is not None


def truncate_filename(filename, max_chars, split_before_max=True):
    """
        Splits a filename into words and truncates it to max_chars.
        If split_before_max is False, it will include the first word that exceeds max_chars.
    """
    words = filename.split()
    result = []
    current_length = 0

    for word in words:
        extra = 1 if result else 0

        if current_length + extra + len(word) > max_chars:
            if not result:
                result.append(word[:max_chars])

            if not split_before_max:
                result.append(" ")
                result.append(word)
            break
        
        if extra:
            result.append(" ")
        result.append(word)
        current_length += extra + len(word)

    return "".join(result)

# TODO: Add to settings, Trunked filename length, split_before_max
def read_czi(path):
    """
        Loads a .czi file and return the data in a proper callable format.
        Made because I could not get a direct reader to work with napari.

        Parameters:
            path: str -> Path to the .czi file.

        Returns:
            callable -> A callable that returns a list of tuples with the data, metadata and layer type.
                        Required format for napari readers.
    """
    # If bfio is present in the environment, show a clear warning now
    # (when the user attempts to open a .czi) so the napari notification
    # system can display it, and then raise to stop the reader from
    # attempting to import problematic readers.
    if BFIO_PRESENT:
        msg = (
            "The installed package `bfio` is known to break CZI reading in this "
            "environment (circular import in `bfio`).\nPlease remove `bfio` from "
            "this environment or use a separate virtual environment without `bfio`."
        )
        try:
            show_warning(msg)
        except Exception:
            import sys
            print(msg, file=sys.stderr)
        raise RuntimeError(msg)

    try:
        from aicsimageio.readers import CziReader
    except Exception as e:
        try:
            show_warning(f"Failed to import `aicsimageio.readers.CziReader`: {e}")
        except Exception:
            import sys
            print(f"Failed to import `aicsimageio.readers.CziReader`: {e}", file=sys.stderr)
        raise

    reader = CziReader(path)
    file_name = os.path.basename(path)
    channels = reader.dims.C
    try:
        metadata_list = extract_key_metadata(reader, channels)
    except ValueError as e:
        show_warning(f"Error extracting metadata from .czi | Is this an Airy scan? | {e}")

        metadata_list = [{} for _ in range(channels)]

    file_name_trunked = truncate_filename(file_name, 20)

    layer_data_list = []
    for channel in range(channels):

        data = reader.get_image_data("ZYX", T=0, C=channel)

        metadata = metadata_list[channel]


        if channels > 1:
            try:
                metadata["name"] = f"{int(float(metadata['metadata']['EmissionWavelength']))}λ | {file_name_trunked}"
            except KeyError:
                metadata["name"] = f"Channel {channel} | {file_name_trunked}"

        if not isinstance(metadata, dict): # Holy shit, I'm making errors
            raise ValueError(f"Metadata for channel {channel} is not a dictionary. Got {type(metadata)}")
        layer_data_list.append((data, metadata, "image"))

    def _reader_callable(_path=None):
        # Napari expect a tuple -> (data, metadata, layer_type)
        # For multiple layers, napari can also take list[tuple] -> [(data, metadata, layer_type)]
        return layer_data_list


    return _reader_callable


# python
from typing import overload
from uuid import UUID


class MultiKeyDict:
    def __init__(self, key_names: tuple = ("name", "unique_id", "wavelength")):
        self._by_name = {}
        self._by_uuid = {}
        self._by_wavelength = {}
        self._keys = key_names

    def __setitem__(self, keys, value):
        # Expecting keys as a tuple of (name, unique_id, wavelength)
        name, unique_id, wavelength = keys
        if isinstance(value, dict):
            value["name"] = name
            value["unique_id"] = unique_id
            value["wavelength"] = wavelength
        self._by_name[name] = value
        self._by_uuid[unique_id] = value
        self._by_wavelength[wavelength] = value


    def __getitem__(self, key):
        # Try to look up the key in all dictionaries.
        if key in self._by_name:
            return self._by_name[key]
        if key in self._by_uuid:
            return self._by_uuid[key]
        if key in self._by_wavelength:
            return self._by_wavelength[key]
        raise KeyError(f"Key {key} not found.")

    def get_all(self,key, value):
        # Return all dicts that have the key-value pair.

        return [entry for entry in self._by_name.values() if entry.get(key, None) == value]


    def has_id(self, uuid: UUID) -> bool:
        return uuid in self._by_uuid


    def to_dict(self):
        return self._by_uuid

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    @overload
    def remove(self, key: str | UUID | int):
        pass
    @overload
    def remove(self, key_list: list[str | UUID | int]):
        pass

    def remove(self, key):
        # If key is a list, remove each item in the list.
        if isinstance(key, list):
            for k in key:
                self.remove(k)
            return

        # Remove the entry from all 3 dictionaries.
        if key in self._by_name:
            uuid = self._by_name[key].get("unique_id")
            wavelength = self._by_name[key].get("wavelength")
            del self._by_name[key]
            del self._by_uuid[uuid]
            del self._by_wavelength[wavelength]
            return

        if key in self._by_uuid:
            name = self._by_uuid[key].get("name")
            wavelength = self._by_uuid[key].get("wavelength")
            del self._by_uuid[key]
            del self._by_name[name]
            del self._by_wavelength[wavelength]
            return

        if key in self._by_wavelength:
            name = self._by_wavelength[key].get("name")
            uuid = self._by_wavelength[key].get("unique_id")
            del self._by_wavelength[key]
            del self._by_name[name]
            del self._by_uuid[uuid]
            return



    def clear(self):
        self._by_name.clear()
        self._by_uuid.clear()
        self._by_wavelength.clear()

    def dump(self, key: UUID | str) -> str:
        # Check if the key is a str. Dump has 2 "modes" depending on key type.
        if isinstance(key, str):
            # In case the key is name, uuid or wavelength, return the corresponding dictionaries.
            result = f"{key} dict:\n"
            # Get the corresponding dictionary.
            _dict = {"name": self._by_name, "unique_id": self._by_uuid, "uuid": self._by_uuid, "wavelength": self._by_wavelength}[key]
            for key, inner_dict in _dict.items():
                # Make a copy so we don't alter the original dictionary.
                inner_copy = inner_dict.copy()
                result += f"{key}: {inner_copy}\n"
            return result


        # Retrieve the entry using the UUID.
        by_uuid_entry = self._by_uuid.get(key)
        if by_uuid_entry is None:
            return f"UUID {key} not found."

        # Get the associated name and wavelength.
        name = by_uuid_entry.get("name")
        wavelength = by_uuid_entry.get("wavelength")

        # Retrieve the corresponding dictionaries.
        by_name_entry = self._by_name.get(name, {})
        by_wavelength_entry = self._by_wavelength.get(wavelength, {})

        result = (
            f"UUID entry: {by_uuid_entry}\n"
            f"Name entry: {by_name_entry}\n"
            f"Wavelength entry: {by_wavelength_entry}\n"
        )
        return result

    def __repr__(self):
        # Return a representation based on one of the dictionaries.
        return f"MultiKeyDict({self._by_name})"

    def __len__(self):
        return len(self._by_name)

    def __next__(self):
        for key in self._by_name:
            yield key

    def __iter__(self):
        return self.__next__()



